/*
    Copyright (C) 2003, Patrick Ahlbrecht

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <getopt.h>
#include <stdlib.h>
#include <libgen.h>
#include <string.h>

#include "util.h"

// Set program version               
char* version=VERSION;

/**
 * Embedding string
 */
char* frmstr;

/**
 * What to use for padding
 */
char pad='0';

/**
 * If the user wishes to map...
 */
char** map=NULL;

/**
 * Index counter of the map. Initially set to the elementcount in map.
 */
int mapindex=-1;

/**
 * Description of the numberbases and the longest string, itos() may produce
 * for the according base
 */
struct basedesc {
  char* name;		// How to refer to this numberbase
  char* digits;		// base description for itos()
  int max;		// Max length from itos() in given range (for padding)
};

/**
 * For macro mappings
 */
struct macromap {
  char* rname;		// Real name
  char* mname;		// Map name
};

struct basedesc* desc;
struct macromap* macros;

/**
 * Lookup macro mapping
 * @param name real name of the macro
 * @return mapped name of the macro or NULL if not found;
 */
const char* getMacro(const char* name) {
  int i=0;
  
  if (macros==NULL) return NULL;
  while(macros[i].rname!=NULL) {
    if (!strcasecmp(macros[i].rname,name)) return macros[i].mname;
    i++;
  }
  return NULL;
}

/**
 * Print usage and exit
 * @param progname argv[0] of main
 */
void printUsage(const char* progname) {
  char* tmp=basename(cpyStr((char*)progname));
  
  printf("%s [OPTION] ...\n\
  \nGenerate sequenced numberstrings\n\n\
  -h\t\t\t\tPrint usage\n\
  -n\t\t\t\tNo sign, use complement for negative numbers\n\
  -c <from>,[<to>[,<step>]]\tSelect counting range and step width\n\
  -t <string>\t\t\tOutput template (default: \"=dec\\n\")\n\
  -g <file>\t\t\tGenerate output, using <file> as template\n\
  -u <string>\t\t\tUse <string> as userdefined numberbase\n\
  -m <string> [<string> ...]\tMap <string>s to numbers (use with \"=map\")\n\
  -p <char>[,<int>]\t\tPadding character [number of leading pads]\n\
  -r <macro>[,<newmacro>]\tRename <macro> to <newmacro>\n\
  \nThe following macros are expanded in the output template:\n\n\
  =bin\tNumber in binary notation\n\
  =oct\tNumber in octal notation\n\
  =dec\tNumber in decimal notation\n\
  =hex\tNumber in hexadecimal notation\n\
  =usr\tNumber in userdefined notation (use with -u switch)\n\
  =map\tString mapping for number (use with -m switch)\n\
  \\n\tNewline character\n\
  \\t\tTab character\n\
  \n",tmp);
  
  exit(EXIT_SUCCESS);
}

/**
 * Actually do the work
 * @param from start value
 * @param to stop value
 * @param step stepping value
 * @param padlen padding length. A value of "-1" means no padding, a value
 * of "0" does autopadding (=all numbers are padded to the stringlength of
 * the biggest number), any value above one will put this number of leading
 * padding characters in front of the biggest number and autopads smaller
 * numbers. NOTE: Biggest number == number with the most digits.
 * @param sign sign negative numbers (only if padlen==-1).
 */
void printString(int from, int to, int step, int padlen, int sign) {
  int x;
  int y;
  char* s[4];
  char* t[4];
  char* tmp=NULL;
  
  for(x=from; ((step<0) && x>=to) || ((step>0) && x<=to)   ;x=x+step) {
    tmp=cpyStr(frmstr);
    y=0;
    while(desc[y].digits!=NULL) {
      if (padlen==-1) {
        
        if (sign && x<0) {
          s[y]=itos(abs(x),desc[y].digits);
          t[y]=(char*)malloc((strlen(s[y])+2)*sizeof(char));
          oomtest(t[y]);
          memcpy(t[y]+1,s[y],strlen(s[y])+1);
          t[y][0]='-';
        }
        else {
          s[y]=itos(x,desc[y].digits);
          t[y]=cpyStr(s[y]);
        }
        
      }
      else {
        s[y]=itos(x,desc[y].digits);
        t[y]=(char*)malloc((desc[y].max+1+padlen)*sizeof(char));
        oomtest(t[y]);
        memset(t[y],pad,desc[y].max+padlen);
        memcpy(t[y]+((padlen+desc[y].max)-strlen(s[y])),s[y],strlen(s[y])+sizeof(char));
      }
      
      tmp=expand(tmp,getMacro(desc[y].name),t[y]);
      
      
      free(s[y]);
      free(t[y]);
      y++;
    }
    
    if(mapindex>=0) {
      tmp=expand(tmp,getMacro("=map"),map[mapindex]);
      mapindex--;
    } else tmp=expand(tmp,getMacro("=map"),NULL);
      
    tmp=expand(tmp,getMacro("\\n"),"\n");
    tmp=expand(tmp,getMacro("\\t"),"\t");
    tmp=expand(tmp,getMacro("=usr"),NULL); // Kill unexpanded macro
        
    printf("%s",tmp);
    free(tmp);
  }
}


int main(int argc, char** argv) {
  int ch;
  int from=0;
  int to=100;
  int padlen=-1;
  int sign=TRUE;
  int step=1;
  
  //frmstr="=dec\\n";
  frmstr=getenv("FSTRING");
  if (frmstr==NULL) frmstr="=dec\\n";
  desc=(struct basedesc*)malloc(7*sizeof(struct basedesc));
  oomtest(desc);
  
  desc[0].name="=bin";
  desc[0].digits="01";
  desc[1].name="=oct";
  desc[1].digits="01234567";
  desc[2].name="=dec";
  desc[2].digits="0123456789";
  desc[3].name="=hex";
  desc[3].digits="0123456789ABCDEF";
  desc[4].name="=usr";
  desc[4].digits=NULL;
  desc[5].name=NULL;
  desc[5].digits=NULL;
  
  macros=(struct macromap*)malloc(10*sizeof(struct macromap));
  oomtest(macros);
  
  macros[0].rname="=bin";
  macros[0].mname="=bin";
  macros[1].rname="=oct";
  macros[1].mname="=oct";
  macros[2].rname="=dec";
  macros[2].mname="=dec";
  macros[3].rname="=hex";
  macros[3].mname="=hex";
  macros[4].rname="=usr";
  macros[4].mname="=usr";
  macros[5].rname="=map";
  macros[5].mname="=map";
  macros[6].rname="\\n";
  macros[6].mname="\\n";
  macros[7].rname="\\t";
  macros[7].mname="\\t";
  macros[8].rname=NULL;
  macros[8].mname=NULL;

  
  while ((ch = getopt(argc, argv, "hr:m:u:t:g:p:c:n")) !=-1) {

    if(map!=NULL) {
      printf("\"-m\" must be the last option!\n");
      return EXIT_FAILURE;
    }
    
    switch (ch) {
      case 'h': {		// Print Usage information
        printUsage(argv[0]);
        break;
      }
      
      case 'c': {		// Range selection
        char** tmp;
        tmp=splitString(optarg,-1,',');
        if (tmp[0]!=NULL) { from=atoi(tmp[0]); to=atoi(tmp[0]); free(tmp[0]);}
        if (tmp[1]!=NULL) { to=atoi(tmp[1]); free(tmp[1]);}
        if (tmp[2]!=NULL) { step=atoi(tmp[2]); free(tmp[2]);}
        break;
      }
      
      case 'u': {		// Base selector
        desc[4].digits=cpyStr(optarg);
        desc[4].digits=optarg;
        break;
      }
      
      case 'm': {		// Unify
        int i;
        for (i=1;i<argc;i++) if (!strcasecmp(argv[i],"-m")) break;
        map=(char**)calloc((argc+1),sizeof(char*));
        oomtest(map);
        memcpy(map,argv+i+1,(argc-i)*sizeof(char*));
        mapindex=argc-i-2;
        to=mapindex*step+from;
        break;
      }
      
      case 't': {		// Template string
        frmstr=cpyStr(optarg);
        frmstr=optarg;
        break;
      }
      
      case 'g': {		// template from file
        frmstr=readFile(optarg);
        if (frmstr==NULL) {
          perror(optarg);
          return EXIT_FAILURE;
        }
        break;
      }
      
      case 'p': {		// Padding character
        char** tmp=splitString(optarg,-1,',');
        pad=optarg[0];
        if(tmp[1]!=NULL) {padlen=atoi(tmp[1]); free(tmp[0]);free(tmp[1]);}
        else padlen=0;
        break;
      }
      
      case 'n': {
        sign=!sign;
        break;
      }
      
      case 'r': {
        char **tmp=splitString(optarg,1,',');
        int i=0;
        while(macros[i].rname!=NULL) {
          if (!strcasecmp(macros[i].rname,tmp[0])) {
            macros[i].mname=(char*)tmp[1];
            break;
          }
          i++;
        }
        free(tmp[0]);
        break;
      }
      
      default: {
        printUsage(argv[0]);
        return EXIT_SUCCESS;
      }
    }
  }
  
  ch=0;
  while(desc[ch].digits!=NULL) {
    char* tmp=itos(from,desc[ch].digits);
    desc[ch].max=strlen(tmp);
    free(tmp);
    tmp=itos(to,desc[ch].digits);
    if (strlen(tmp)>desc[ch].max) desc[ch].max=strlen(tmp);
    free(tmp);
    ch++;
    
  }

  printString(from,to,step,padlen,sign);
  return 0;
}
